# --------------------------------------------------------------------------------------
# Copyright (c) 2021-2023, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
from typing import Any, List as TList, Optional, Tuple, Type, TypeVar, overload

from .catom import Member

T = TypeVar("T")
T1 = TypeVar("T1")
T2 = TypeVar("T2")

class List(Member[TList[T], TList[T]]):
    # No default
    @overload
    def __new__(
        cls, kind: None = None, default: Optional[TList[Any]] = None
    ) -> List[Any]: ...
    @overload
    def __new__(cls, kind: Type[T], default: None = None) -> List[T]: ...
    @overload
    def __new__(cls, kind: Tuple[Type[T]], default: None = None) -> List[T]: ...
    @overload
    def __new__(
        cls, kind: Tuple[Type[T], Type[T1]], default: None = None
    ) -> List[T | T1]: ...
    @overload
    def __new__(
        cls,
        kind: Tuple[Type[T], Type[T1], Type[T2]],
        default: None = None,
    ) -> List[T | T1 | T2]: ...
    @overload
    def __new__(cls, kind: Member[T, Any], default: None = None) -> List[T]: ...
    # With default
    @overload
    def __new__(cls, kind: Type[T], default: TList[T]) -> List[T]: ...
    @overload
    def __new__(cls, kind: Tuple[Type[T]], default: TList[T]) -> List[T]: ...
    @overload
    def __new__(
        cls, kind: Tuple[Type[T], Type[T1]], default: TList[T | T1]
    ) -> List[T | T1]: ...
    @overload
    def __new__(
        cls, kind: Tuple[Type[T], Type[T1]], default: TList[T] | TList[T1]
    ) -> List[T | T1]: ...
    @overload
    def __new__(
        cls, kind: Tuple[Type[T], Type[T1], Type[T2]], default: TList[T | T1 | T2]
    ) -> List[T | T1 | T2]: ...
    @overload
    def __new__(
        cls,
        kind: Tuple[Type[T], Type[T1], Type[T2]],
        default: TList[T | T1] | TList[T | T2] | TList[T1 | T2],
    ) -> List[T | T1 | T2]: ...
    @overload
    def __new__(
        cls,
        kind: Tuple[Type[T], Type[T1], Type[T2]],
        default: TList[T] | TList[T1] | TList[T2],
    ) -> List[T | T1 | T2]: ...
    @overload
    def __new__(cls, kind: Member[T, Any], default: TList[T]) -> List[T]: ...



# --------------------------------------------------------------------------------------
# Copyright (c) 2013-2023, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
from .catom import DefaultValue, Member, Validate
from .instance import Instance
from .typing_utils import extract_types, is_optional


class Dict(Member):
    """A value of type `dict`."""

    __slots__ = ()

    def __init__(self, key=None, value=None, default=None):
        """Initialize a Dict.

        Parameters
        ----------
        key : Member, type, tuple of types, or None, optional
            A member to use for validating the types of keys allowed in
            the dict. This can also be a type or a tuple of types, which
            will be wrapped with an Instance member. If this is not
            given, no key validation is performed.

        value : Member, type, tuple of types, or None, optional
            A member to use for validating the types of values allowed
            in the dict. This can also be a type or a tuple of types,
            which will be wrapped with an Instance member. If this is
            not given, no value validation is performed.

        default : dict, optional
            The default dict of items. A new copy of this dict will be
            created for each atom instance.

        """
        self.set_default_value_mode(DefaultValue.Dict, default)
        if key is not None and not isinstance(key, Member):
            opt, types = is_optional(extract_types(key))
            key = Instance(types, optional=opt)
        if value is not None and not isinstance(value, Member):
            opt, types = is_optional(extract_types(value))
            value = Instance(types, optional=opt)
        self.set_validate_mode(Validate.Dict, (key, value))

    def set_name(self, name):
        """Assign the name to this member.

        This method is called by the Atom metaclass when a class is
        created. This makes sure the name of the internal members are
        also updated.

        """
        super(Dict, self).set_name(name)
        key, value = self.validate_mode[1]
        if key is not None:
            key.set_name(name + "|key")
        if value is not None:
            value.set_name(name + "|value")

    def set_index(self, index):
        """Assign the index to this member.

        This method is called by the Atom metaclass when a class is
        created. This makes sure the index of the internal members are
        also updated.

        """
        super(Dict, self).set_index(index)
        key, value = self.validate_mode[1]
        if key is not None:
            key.set_index(index)
        if value is not None:
            value.set_index(index)

    def clone(self):
        """Create a clone of the member.

        This will clone the internal dict key and value members if they exist.

        """
        clone = super(Dict, self).clone()
        key, value = self.validate_mode[1]
        if key is not None or value is not None:
            key_clone = key.clone() if key is not None else None
            value_clone = value.clone() if value is not None else None
            mode, _ = self.validate_mode
            clone.set_validate_mode(mode, (key_clone, value_clone))
        return clone



# --------------------------------------------------------------------------------------
# Copyright (c) 2013-2023, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
from .catom import (
    DefaultValue,
    Member,
    PostGetAttr,
    PostSetAttr,
    PostValidate,
    Validate,
)


class Delegator(Member):
    """A member subclass which delegates all work to a wrapped member.

    The only behaviors not delegated are GetAttr and SetAttr. Subclasses
    should override behavior as needed to suit their needs. In order to
    change a particular mode, the relevant change method must be called
    via super(Delegator, ...).

    """

    __slots__ = "delegate"

    def __init__(self, delegate):
        """Initialize a DeclarativeProperty.

        Parameters
        ----------
        delegate : Member
            The Atom Member which provides the behavior for the property.
            The member should use standard slot behavior semantics.

        """
        self.delegate = delegate
        sup = super(Delegator, self)
        sup.set_post_getattr_mode(PostGetAttr.Delegate, delegate)
        sup.set_post_setattr_mode(PostSetAttr.Delegate, delegate)
        sup.set_default_value_mode(DefaultValue.Delegate, delegate)
        sup.set_validate_mode(Validate.Delegate, delegate)
        sup.set_post_validate_mode(PostValidate.Delegate, delegate)

    def add_static_observer(self, observer):
        """Add a static observer to the member.

        This method also adds the static observer to the delegate.

        """
        super(Delegator, self).add_static_observer(observer)
        self.delegate.add_static_observer(observer)

    def remove_static_observer(self, observer):
        """Remove a static observer from the member.

        This method also removes the static observer from the delegate.

        """
        super(Delegator, self).remove_static_observer(observer)
        self.delegate.remove_static_observer(observer)

    def set_name(self, name):
        """Assign the name to this member.

        This method keeps the name of the delegate member in sync.

        """
        super(Delegator, self).set_name(name)
        self.delegate.set_name(name)

    def set_index(self, index):
        """Assign the index to this member.

        This method keeps the index of the delegate member in sync.

        """
        super(Delegator, self).set_index(index)
        self.delegate.set_index(index)

    def set_post_getattr_mode(self, mode, context):
        """Set the post getattr mode for the member.

        This method proxies the change to the delegate member.

        """
        self.delegate.set_post_getattr_mode(mode, context)

    def set_post_setattr_mode(self, mode, context):
        """Set the post getattr mode for the member.

        This method proxies the change to the delegate member.

        """
        self.delegate.set_post_setattr_mode(mode, context)

    def set_default_value_mode(self, mode, context):
        """Set the default value mode for the member.

        This method proxies the change to the delegate member.

        """
        self.delegate.set_default_value_mode(mode, context)

    def set_validate_mode(self, mode, context):
        """Set the default value mode for the member.

        This method proxies the change to the delegate member.

        """
        self.delegate.set_validate_mode(mode, context)

    def set_post_validate_mode(self, mode, context):
        """Set the default value mode for the member.

        This method proxies the change to the delegate member.

        """
        self.delegate.set_post_validate_mode(mode, context)

    def clone(self):
        """Create a clone of the declarative property.

        This method also creates a clone of the internal delegate for
        mode handlers which use the original delegate as the context.

        """
        clone = super(Delegator, self).clone()
        delegate = self.delegate
        clone.delegate = delegate_clone = delegate.clone()

        mode, old = clone.post_getattr_mode
        if old is delegate:
            clone.set_post_getattr_mode(mode, delegate_clone)

        mode, old = clone.post_setattr_mode
        if old is delegate:
            clone.set_post_setattr_mode(mode, delegate_clone)

        mode, old = clone.default_value_mode
        if old is delegate:
            clone.set_default_value_mode(mode, delegate_clone)

        mode, old = clone.validate_mode
        if old is delegate:
            clone.set_validate_mode(mode, delegate_clone)

        mode, old = clone.post_validate_mode
        if old is delegate:
            clone.set_post_validate_mode(mode, delegate_clone)

        return clone



# --------------------------------------------------------------------------------------
# Copyright (c) 2021-2023, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
from enum import IntEnum, IntFlag
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Literal,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    overload,
)

from typing_extensions import Self

from .atom import Atom
from .property import Property
from .typing_utils import ChangeDict

def reset_property(prop: Property[Any, Any], owner: Atom) -> None: ...

class CAtom:
    def __init__(self, **kwargs: Any) -> None: ...
    def freeze(self) -> None: ...
    def get_member(self, member: str) -> Member[Any, Any]: ...
    def has_observer(
        self, member: str, func: Callable[[Dict[str, Any]], None]
    ) -> bool: ...
    def has_observers(self, member: str) -> bool: ...
    def notifications_enabled(self) -> bool: ...
    def notify(self, member_name: str, *args: Any, **kwargs: Any) -> None: ...
    def observe(
        self,
        member: str,
        func: Callable[[ChangeDict], None],
        change_types: ChangeType = ChangeType.ANY,
    ) -> None: ...
    def set_notifications_enabled(self, enabled: bool) -> bool: ...
    def unobserve(self, member: str, func: Callable[[ChangeDict], None]) -> None: ...
    def __sizeof__(self) -> int: ...

T = TypeVar("T")
S = TypeVar("S")

class Member(Generic[T, S]):
    index: int = ...
    metadata: Optional[Dict[str, Any]] = ...
    name: str = ...
    # Mode accessors
    default_value_mode: Tuple[DefaultValue, Any] = ...
    delattr_mode: Tuple[DelAttr, Any] = ...
    getattr_mode: Tuple[GetAttr, Any] = ...
    post_getattr_mode: Tuple[PostGetAttr, Any] = ...
    post_setattr_mode: Tuple[PostSetAttr, Any] = ...
    post_validate_mode: Tuple[PostValidate, Any] = ...
    setattr_mode: Tuple[SetAttr, Any] = ...
    validate_mode: Tuple[Validate, Any] = ...
    getstate_mode: Tuple[GetState, Any] = ...
    def __init__(self) -> None: ...
    @overload
    def __get__(self, instance: None, owner: Type[Atom]) -> Self: ...
    @overload
    def __get__(self, instance: Atom, owner: Type[Atom]) -> T: ...
    def __set__(self, instance: Atom, value: S) -> None: ...
    def __delete__(self, instance: Atom) -> None: ...
    def tag(self, **kwargs: Any) -> Self: ...
    def clone(self) -> Self: ...
    def add_static_observer(
        self,
        observer: str | Callable[[ChangeDict], None],
        change_types: ChangeType = ChangeType.ANY,
    ) -> Any: ...
    def remove_static_observer(
        self, observer: str | Callable[[ChangeDict], None]
    ) -> Any: ...
    def static_observers(self) -> Tuple[str | Callable[[ChangeDict], None], ...]: ...
    def has_observer(
        self,
        observer: str | Callable[[ChangeDict], None],
        change_types: ChangeType = ChangeType.ANY,
    ) -> Any: ...
    def has_observers(self, change_types: ChangeType = ChangeType.ANY) -> bool: ...
    def copy_static_observers(self, other: Member[Any, Any]) -> None: ...
    def notify(self, owner: CAtom, *args: Any, **kwargs: Any) -> None: ...
    def get_slot(self, owner: CAtom) -> Any: ...
    def set_slot(self, owner: CAtom, value: Any) -> None: ...
    def del_slot(self, owner: CAtom) -> None: ...
    # Manual operation of the member
    def do_default_value(self, owner: CAtom) -> Any: ...
    def do_delattr(self, owner: CAtom) -> None: ...
    def do_full_validate(self, owner: CAtom, old: T, new: Any) -> T: ...
    def do_getattr(self, owner: CAtom) -> Any: ...
    def do_post_getattr(self, owner: CAtom, value: Any) -> T: ...
    def do_post_setattr(self, owner: CAtom, old: T, new: T) -> None: ...
    def do_post_validate(self, owner: CAtom, old: T, new: T) -> T: ...
    def do_setattr(self, owner: CAtom, value: Any) -> Any: ...
    def do_validate(self, owner: CAtom, old: T, new: Any) -> T: ...
    def do_should_getstate(self, owner: CAtom) -> bool: ...
    # Setter for the member
    def set_index(self, index: int) -> None: ...
    def set_name(self, name: str) -> None: ...
    # Default value mode
    @overload
    def set_default_value_mode(
        self, mode: Literal[DefaultValue.List], context: list[Any]
    ) -> None: ...
    @overload
    def set_default_value_mode(
        self, mode: Literal[DefaultValue.Set], context: set[Any]
    ) -> None: ...
    @overload
    def set_default_value_mode(
        self, mode: Literal[DefaultValue.Dict], context: dict[Any, Any]
    ) -> None: ...
    @overload
    def set_default_value_mode(
        self, mode: Literal[DefaultValue.Delegate], context: Member[T, S]
    ) -> None: ...
    @overload
    def set_default_value_mode(
        self, mode: Literal[DefaultValue.CallObject], context: Callable[[], Any]
    ) -> None: ...
    @overload
    def set_default_value_mode(
        self,
        mode: Literal[DefaultValue.CallObject_Object],
        context: Callable[[CAtom], Any],
    ) -> None: ...
    @overload
    def set_default_value_mode(
        self,
        mode: Literal[DefaultValue.CallObject_ObjectName],
        context: Callable[[CAtom, str], Any],
    ) -> None: ...
    @overload
    def set_default_value_mode(
        self,
        mode: Literal[DefaultValue.ObjectMethod]
        | Literal[DefaultValue.ObjectMethod_Name]
        | Literal[DefaultValue.MemberMethod_Object],
        context: str,
    ) -> None: ...
    @overload
    def set_default_value_mode(
        self, mode: Literal[DefaultValue.Static], context: Any
    ) -> None: ...
    @overload
    def set_default_value_mode(
        self, mode: Literal[DefaultValue.NonOptional], context: None
    ) -> None: ...
    @overload
    def set_default_value_mode(
        self, mode: Literal[DefaultValue.NoOp], context: None
    ) -> None: ...
    # Delattr mode
    @overload
    def set_delattr_mode(
        self, mode: Literal[DelAttr.Delegate], context: Member[T, S]
    ) -> None: ...
    @overload
    def set_delattr_mode(
        self,
        mode: Literal[DelAttr.Property],
        context: Optional[Callable[[CAtom], None]],
    ) -> None: ...
    @overload
    def set_delattr_mode(
        self,
        mode: Literal[DelAttr.Constant]
        | Literal[DelAttr.Event]
        | Literal[DelAttr.NoOp]
        | Literal[DelAttr.ReadOnly]
        | Literal[DelAttr.Signal]
        | Literal[DelAttr.Slot],
        context: None,
    ) -> None: ...
    # Getattr mode
    @overload
    def set_getattr_mode(
        self, mode: Literal[GetAttr.Delegate], context: Member[T, S]
    ) -> None: ...
    @overload
    def set_getattr_mode(
        self,
        mode: Literal[GetAttr.Property] | Literal[GetAttr.CachedProperty],
        context: Optional[Callable[[CAtom], Any]],
    ) -> None: ...
    @overload
    def set_getattr_mode(
        self, mode: Literal[GetAttr.CallObject_Object], context: Callable[[CAtom], Any]
    ) -> None: ...
    @overload
    def set_getattr_mode(
        self,
        mode: Literal[GetAttr.CallObject_ObjectName],
        context: Callable[[CAtom, str], Any],
    ) -> None: ...
    @overload
    def set_getattr_mode(
        self,
        mode: Literal[GetAttr.ObjectMethod]
        | Literal[GetAttr.ObjectMethod_Name]
        | Literal[GetAttr.MemberMethod_Object],
        context: str,
    ) -> None: ...
    @overload
    def set_getattr_mode(
        self,
        mode: Literal[GetAttr.Event]
        | Literal[GetAttr.NoOp]
        | Literal[GetAttr.Signal]
        | Literal[GetAttr.Slot],
        context: str,
    ) -> None: ...
    # Post getattr mode
    @overload
    def set_post_getattr_mode(
        self, mode: Literal[PostGetAttr.NoOp], context: None
    ) -> None: ...
    @overload
    def set_post_getattr_mode(
        self, mode: Literal[PostGetAttr.Delegate], context: Member[T, S]
    ) -> None: ...
    @overload
    def set_post_getattr_mode(
        self,
        mode: Literal[PostGetAttr.ObjectMethod_Value]
        | Literal[PostGetAttr.ObjectMethod_NameValue]
        | Literal[PostGetAttr.MemberMethod_ObjectValue],
        context: str,
    ) -> None: ...
    # Post setattr mode
    @overload
    def set_post_setattr_mode(
        self, mode: Literal[PostSetAttr.NoOp], context: None
    ) -> None: ...
    @overload
    def set_post_setattr_mode(
        self, mode: Literal[PostSetAttr.Delegate], context: Member[T, S]
    ) -> None: ...
    @overload
    def set_post_setattr_mode(
        self,
        mode: Literal[PostSetAttr.ObjectMethod_OldNew]
        | Literal[PostSetAttr.ObjectMethod_NameOldNew]
        | Literal[PostSetAttr.MemberMethod_ObjectOldNew],
        context: str,
    ) -> None: ...
    # Post validate mode
    @overload
    def set_post_validate_mode(
        self, mode: Literal[PostValidate.NoOp], context: None
    ) -> None: ...
    @overload
    def set_post_validate_mode(
        self, mode: Literal[PostValidate.Delegate], context: Member[T, S]
    ) -> None: ...
    @overload
    def set_post_validate_mode(
        self,
        mode: Literal[PostValidate.ObjectMethod_OldNew]
        | Literal[PostValidate.ObjectMethod_NameOldNew]
        | Literal[PostValidate.MemberMethod_ObjectOldNew],
        context: str,
    ) -> None: ...
    # Setattr mode
    @overload
    def set_setattr_mode(
        self,
        mode: Literal[SetAttr.Constant]
        | Literal[SetAttr.Event]
        | Literal[SetAttr.NoOp]
        | Literal[SetAttr.ReadOnly]
        | Literal[SetAttr.Signal]
        | Literal[SetAttr.Slot],
        context: None,
    ) -> None: ...
    @overload
    def set_setattr_mode(
        self, mode: Literal[SetAttr.Delegate], context: Member[T, S]
    ) -> None: ...
    @overload
    def set_setattr_mode(
        self,
        mode: Literal[SetAttr.Property],
        context: Optional[Callable[[CAtom], Any]],
    ) -> None: ...
    @overload
    def set_setattr_mode(
        self,
        mode: Literal[SetAttr.CallObject_ObjectValue],
        context: Callable[[CAtom, Any], Any],
    ) -> None: ...
    @overload
    def set_setattr_mode(
        self,
        mode: Literal[SetAttr.CallObject_ObjectNameValue],
        context: Callable[[CAtom, str, Any], Any],
    ) -> None: ...
    @overload
    def set_setattr_mode(
        self,
        mode: Literal[SetAttr.ObjectMethod_Value]
        | Literal[SetAttr.ObjectMethod_NameValue]
        | Literal[SetAttr.MemberMethod_ObjectValue],
        context: str,
    ) -> None: ...
    # Validate mode
    @overload
    def set_validate_mode(
        self,
        mode: Literal[Validate.Bool]
        | Literal[Validate.Bytes]
        | Literal[Validate.BytesPromote]
        | Literal[Validate.Callable]
        | Literal[Validate.Float]
        | Literal[Validate.FloatPromote]
        | Literal[Validate.Int]
        | Literal[Validate.IntPromote]
        | Literal[Validate.NoOp]
        | Literal[Validate.Str]
        | Literal[Validate.StrPromote],
        context: None,
    ) -> None: ...
    @overload
    def set_validate_mode(
        self,
        mode: Literal[Validate.Tuple]
        | Literal[Validate.List]
        | Literal[Validate.ContainerList]
        | Literal[Validate.Set],
        context: Optional[Member[Any, Any]],
    ) -> None: ...
    @overload
    def set_validate_mode(
        self,
        mode: Literal[Validate.Dict],
        context: Tuple[Optional[Member[Any, Any]], Optional[Member[Any, Any]]],
    ) -> None: ...
    @overload
    def set_validate_mode(
        self,
        mode: Literal[Validate.Instance]
        | Literal[Validate.OptionalInstance]
        | Literal[Validate.Subclass],
        context: type | Tuple[type, ...],
    ) -> None: ...
    @overload
    def set_validate_mode(
        self,
        mode: Literal[Validate.Typed] | Literal[Validate.OptionalTyped],
        context: type,
    ) -> None: ...
    @overload
    def set_validate_mode(
        self, mode: Literal[Validate.Enum], context: Sequence[Any]
    ) -> None: ...
    @overload
    def set_validate_mode(
        self,
        mode: Literal[Validate.FloatRange],
        context: Tuple[Optional[float], Optional[float]],
    ) -> None: ...
    @overload
    def set_validate_mode(
        self,
        mode: Literal[Validate.Range],
        context: Tuple[Optional[int], Optional[int]],
    ) -> None: ...
    @overload
    def set_validate_mode(
        self,
        mode: Literal[Validate.Coerced],
        context: Tuple[Type[T], Callable[[Any], T]],
    ) -> None: ...
    @overload
    def set_validate_mode(
        self, mode: Literal[Validate.Delegate], context: Member[T, S]
    ) -> None: ...
    @overload
    def set_validate_mode(
        self,
        mode: Literal[Validate.ObjectMethod_OldNew]
        | Literal[Validate.ObjectMethod_NameOldNew]
        | Literal[Validate.MemberMethod_ObjectOldNew],
        context: str,
    ) -> None: ...
    # Getstate mode
    @overload
    def set_getstate_mode(
        self,
        mode: Literal[GetState.Exclude]
        | Literal[GetState.Include]
        | Literal[GetState.IncludeNonDefault]
        | Literal[GetState.Property],
        context: None,
    ) -> None: ...
    @overload
    def set_getstate_mode(
        self,
        mode: Literal[GetState.MemberMethod_Object]
        | Literal[GetState.ObjectMethod_Name],
        context: str,
    ) -> None: ...

KT = TypeVar("KT")
VT = TypeVar("VT")

class atomlist(List[T]): ...
class atomclist(atomlist[T]): ...
class atomset(Set[T]): ...
class atomdict(Dict[KT, VT]): ...

A = TypeVar("A", bound=CAtom)

class atomref(Generic[A]):
    def __new__(cls, atom: A) -> atomref[A]: ...
    def __bool__(self) -> bool: ...
    def __call__(self) -> Optional[A]: ...
    def __sizeof__(self) -> int: ...

class SignalConnector:
    def __call__(self, *args: Any, **kwargs: Any) -> None: ...
    def emit(self, *args: Any, **kwargs: Any) -> None: ...
    def connect(self, slot: Callable[..., Any]) -> None: ...
    def disconnect(self, slot: Callable[..., Any]) -> None: ...

class EventBinder:
    def bind(self, observer: Callable[[Dict[str, Any]], None]) -> None: ...
    def unbind(self, observer: Callable[[Dict[str, Any]], None]) -> None: ...

class DefaultValue(IntEnum):
    CallObject = ...
    CallObject_Object = ...
    CallObject_ObjectName = ...
    Delegate = ...
    Dict = ...
    List = ...
    MemberMethod_Object = ...
    NonOptional = ...
    NoOp = ...
    ObjectMethod = ...
    ObjectMethod_Name = ...
    Set = ...
    Static = ...

class DelAttr(IntEnum):
    Constant = ...
    Delegate = ...
    Event = ...
    NoOp = ...
    Property = ...
    ReadOnly = ...
    Signal = ...
    Slot = ...

class GetAttr(IntEnum):
    CachedProperty = ...
    CallObject_Object = ...
    CallObject_ObjectName = ...
    Delegate = ...
    Event = ...
    MemberMethod_Object = ...
    NoOp = ...
    ObjectMethod = ...
    ObjectMethod_Name = ...
    Property = ...
    Signal = ...
    Slot = ...

class PostGetAttr(IntEnum):
    Delegate = ...
    MemberMethod_ObjectValue = ...
    NoOp = ...
    ObjectMethod_NameValue = ...
    ObjectMethod_Value = ...

class PostSetAttr(IntEnum):
    Delegate = ...
    MemberMethod_ObjectOldNew = ...
    NoOp = ...
    ObjectMethod_NameOldNew = ...
    ObjectMethod_OldNew = ...

class PostValidate(IntEnum):
    Delegate = ...
    MemberMethod_ObjectOldNew = ...
    NoOp = ...
    ObjectMethod_NameOldNew = ...
    ObjectMethod_OldNew = ...

class SetAttr(IntEnum):
    CallObject_ObjectNameValue = ...
    CallObject_ObjectValue = ...
    Constant = ...
    Delegate = ...
    Event = ...
    MemberMethod_ObjectValue = ...
    NoOp = ...
    ObjectMethod_NameValue = ...
    ObjectMethod_Value = ...
    Property = ...
    ReadOnly = ...
    Signal = ...
    Slot = ...

class Validate(IntEnum):
    Bool = ...
    Bytes = ...
    BytesPromote = ...
    Callable = ...
    Coerced = ...
    ContainerList = ...
    Delegate = ...
    Dict = ...
    Enum = ...
    Float = ...
    FloatPromote = ...
    FloatRange = ...
    Instance = ...
    Int = ...
    IntPromote = ...
    List = ...
    MemberMethod_ObjectOldNew = ...
    OptionalInstance = ...
    OptionalTyped = ...
    NoOp = ...
    ObjectMethod_NameOldNew = ...
    ObjectMethod_OldNew = ...
    Range = ...
    Set = ...
    Str = ...
    StrPromote = ...
    Subclass = ...
    Tuple = ...
    Typed = ...

class GetState(IntEnum):
    Exclude = ...
    Include = ...
    IncludeNonDefault = ...
    Property = ...
    MemberMethod_Object = ...
    ObjectMethod_Name = ...

class ChangeType(IntFlag):
    CREATE = ...
    UPDATE = ...
    DELETE = ...
    EVENT = ...
    PROPERTY = ...
    CONTAINER = ...
    ANY = ...



# --------------------------------------------------------------------------------------
# Copyright (c) 2013-2023, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
"""Module exporting the public interface to atom.

"""
from .atom import Atom, AtomMeta, MissingMemberWarning, add_member, observe, set_default
from .catom import (
    CAtom,
    ChangeType,
    DefaultValue,
    GetAttr,
    GetState,
    Member,
    PostGetAttr,
    PostSetAttr,
    PostValidate,
    SetAttr,
    Validate,
    atomclist,
    atomdict,
    atomlist,
    atomref,
    atomset,
)
from .coerced import Coerced
from .containerlist import ContainerList
from .delegator import Delegator
from .dict import Dict
from .enum import Enum
from .event import Event
from .instance import ForwardInstance, Instance
from .list import List
from .property import Property, cached_property
from .scalars import (
    Bool,
    Bytes,
    Callable,
    Constant,
    Float,
    FloatRange,
    Int,
    Range,
    ReadOnly,
    Str,
    Value,
)
from .set import Set
from .signal import Signal
from .subclass import ForwardSubclass, Subclass
from .tuple import Tuple
from .typed import ForwardTyped, Typed
from .typing_utils import ChangeDict

__all__ = [
    "Atom",
    "AtomMeta",
    "MissingMemberWarning",
    "add_member",
    "observe",
    "set_default",
    "CAtom",
    "ChangeType",
    "DefaultValue",
    "GetAttr",
    "GetState",
    "Member",
    "PostGetAttr",
    "PostSetAttr",
    "PostValidate",
    "SetAttr",
    "Validate",
    "atomclist",
    "atomdict",
    "atomlist",
    "atomref",
    "atomset",
    "Coerced",
    "ContainerList",
    "Delegator",
    "Dict",
    "Enum",
    "Event",
    "ForwardInstance",
    "Instance",
    "List",
    "Property",
    "cached_property",
    "Bool",
    "Bytes",
    "Callable",
    "Constant",
    "Float",
    "FloatRange",
    "Int",
    "Range",
    "ReadOnly",
    "Str",
    "Value",
    "Set",
    "Signal",
    "ForwardSubclass",
    "Subclass",
    "Tuple",
    "ForwardTyped",
    "Typed",
    "ChangeDict",
]



# --------------------------------------------------------------------------------------
# Copyright (c) 2021, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
from typing import Any, List, Optional, Tuple, Type, TypeVar, overload

from .catom import Member

T = TypeVar("T")
T1 = TypeVar("T1")
T2 = TypeVar("T2")

class ContainerList(Member[List[T], List[T]]):
    # No default
    @overload
    def __new__(
        cls, kind: None = None, default: Optional[List[Any]] = None
    ) -> ContainerList[Any]: ...
    @overload
    def __new__(cls, kind: Type[T], default: None = None) -> ContainerList[T]: ...
    @overload
    def __new__(
        cls, kind: Tuple[Type[T]], default: None = None
    ) -> ContainerList[T]: ...
    @overload
    def __new__(
        cls, kind: Tuple[Type[T], Type[T1]], default: None = None
    ) -> ContainerList[T | T1]: ...
    @overload
    def __new__(
        cls,
        kind: Tuple[Type[T], Type[T1], Type[T2]],
        default: None = None,
    ) -> ContainerList[T | T1 | T2]: ...
    @overload
    def __new__(
        cls, kind: Member[T, Any], default: None = None
    ) -> ContainerList[T]: ...
    # With default
    @overload
    def __new__(cls, kind: Type[T], default: List[T]) -> ContainerList[T]: ...
    @overload
    def __new__(cls, kind: Tuple[Type[T]], default: List[T]) -> ContainerList[T]: ...
    @overload
    def __new__(
        cls, kind: Tuple[Type[T], Type[T1]], default: List[T | T1]
    ) -> ContainerList[T | T1]: ...
    @overload
    def __new__(
        cls, kind: Tuple[Type[T], Type[T1]], default: List[T] | List[T1]
    ) -> ContainerList[T | T1]: ...
    @overload
    def __new__(
        cls, kind: Tuple[Type[T], Type[T1], Type[T2]], default: List[T | T1 | T2]
    ) -> ContainerList[T | T1 | T2]: ...
    @overload
    def __new__(
        cls,
        kind: Tuple[Type[T], Type[T1], Type[T2]],
        default: List[T | T1] | List[T | T2] | List[T1 | T2],
    ) -> ContainerList[T | T1 | T2]: ...
    @overload
    def __new__(
        cls,
        kind: Tuple[Type[T], Type[T1], Type[T2]],
        default: List[T] | List[T1] | List[T2],
    ) -> ContainerList[T | T1 | T2]: ...
    @overload
    def __new__(
        cls, kind: Member[T, Any], default: Optional[List[T]] = None
    ) -> ContainerList[T]: ...



# --------------------------------------------------------------------------------------
# Copyright (c) 2021, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
from typing import Any, Callable, Dict, Optional, Tuple, Type, TypeVar, overload

from .catom import Member

T = TypeVar("T")
T1 = TypeVar("T1")
T2 = TypeVar("T2")

S = TypeVar("S")

class Coerced(Member[T, S]):
    # No default
    # - type
    @overload
    def __new__(
        cls,
        kind: Type[T],
        args: Optional[Tuple[Any, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
        *,
        factory: None = None,
        coercer: None = None,
    ) -> Coerced[T, T]: ...
    @overload
    def __new__(
        cls,
        kind: Type[T],
        args: Optional[Tuple[Any, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
        *,
        factory: None = None,
        coercer: Callable[[S], T],
    ) -> Coerced[T, T | S]: ...
    # - 1-Tuple[Any, ...]
    @overload
    def __new__(
        cls,
        kind: Tuple[Type[T]],
        args: Optional[Tuple[Any, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
        *,
        factory: None = None,
        coercer: None = None,
    ) -> Coerced[T, T]: ...
    @overload
    def __new__(
        cls,
        kind: Tuple[Type[T]],
        args: Optional[Tuple[Any, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
        *,
        factory: None = None,
        coercer: Callable[[S], T],
    ) -> Coerced[T, T | S]: ...
    # - 2-Tuple[Any, ...]
    @overload
    def __new__(
        cls,
        kind: Tuple[Type[T], Type[T1]],
        args: Optional[Tuple[Any, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
        *,
        factory: None = None,
        coercer: None = None,
    ) -> Coerced[T | T1, T | T1]: ...
    @overload
    def __new__(
        cls,
        kind: Tuple[Type[T], Type[T1]],
        args: Optional[Tuple[Any, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
        *,
        factory: None = None,
        coercer: Callable[[S], T | T1],
    ) -> Coerced[T | T1, T | T1 | S]: ...
    # - 3-Tuple[Any, ...]
    @overload
    def __new__(
        cls,
        kind: Tuple[Type[T], Type[T1], Type[T2]],
        args: Optional[Tuple[Any, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
        *,
        factory: None = None,
        coercer: None = None,
    ) -> Coerced[T | T1 | T2, T | T1 | T2]: ...
    @overload
    def __new__(
        cls,
        kind: Tuple[Type[T], Type[T1], Type[T2]],
        args: Optional[Tuple[Any, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
        *,
        factory: None = None,
        coercer: Callable[[S], T | T1 | T2],
    ) -> Coerced[T | T1 | T2, T | T1 | T2 | S]: ...
    # Default with factory
    # - type
    @overload
    def __new__(
        cls,
        kind: Type[T],
        args: None = None,
        kwargs: None = None,
        *,
        factory: Callable[[], T],
        coercer: None = None,
    ) -> Coerced[T, T]: ...
    @overload
    def __new__(
        cls,
        kind: Type[T],
        args: None = None,
        kwargs: None = None,
        *,
        factory: Callable[[], T | S],
        coercer: Callable[[S], T],
    ) -> Coerced[T, T | S]: ...
    # - 1-Tuple[Any, ...]
    @overload
    def __new__(
        cls,
        kind: Tuple[Type[T]],
        args: None = None,
        kwargs: None = None,
        *,
        factory: Callable[[], T],
        coercer: None = None,
    ) -> Coerced[T, T]: ...
    @overload
    def __new__(
        cls,
        kind: Tuple[Type[T]],
        args: None = None,
        kwargs: None = None,
        *,
        factory: Callable[[], T | S],
        coercer: Callable[[S], T],
    ) -> Coerced[T, T | S]: ...
    # - 2-Tuple[Any, ...]
    @overload
    def __new__(
        cls,
        kind: Tuple[Type[T], Type[T1]],
        args: None = None,
        kwargs: None = None,
        *,
        factory: Callable[[], T | T1],
        coercer: None = None,
    ) -> Coerced[T | T1, T | T1]: ...
    @overload
    def __new__(
        cls,
        kind: Tuple[Type[T], Type[T1]],
        args: None = None,
        kwargs: None = None,
        *,
        factory: Callable[[], T | T1 | S],
        coercer: Callable[[S], T | T1],
    ) -> Coerced[T | T1, T | T1 | S]: ...
    # - 3-Tuple[Any, ...]
    @overload
    def __new__(
        cls,
        kind: Tuple[Type[T], Type[T1], Type[T2]],
        args: None = None,
        kwargs: None = None,
        *,
        factory: Callable[[], T | T1 | T2],
        coercer: None = None,
    ) -> Coerced[T | T1 | T2, T | T1 | T2]: ...
    @overload
    def __new__(
        cls,
        kind: Tuple[Type[T], Type[T1], Type[T2]],
        args: None = None,
        kwargs: None = None,
        *,
        factory: Callable[[], S],
        coercer: Callable[[S], T | T1 | T2 | S],
    ) -> Coerced[T | T1 | T2, T | T1 | T2 | S]: ...



# --------------------------------------------------------------------------------------
# Copyright (c) 2021-2023, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
from typing import Any, Tuple, Type, TypeVar, Union, overload

from .catom import EventBinder, Member

T = TypeVar("T")
T1 = TypeVar("T1")
T2 = TypeVar("T2")

class Event(Member[EventBinder, T]):
    @overload
    def __new__(cls, kind: None = None) -> Event[Any]: ...
    @overload
    def __new__(cls, kind: Type[T]) -> Event[T]: ...
    @overload
    def __new__(cls, kind: Tuple[Type[T]]) -> Event[T]: ...
    @overload
    def __new__(cls, kind: Tuple[Type[T], Type[T1]]) -> Event[Union[T, T1]]: ...
    @overload
    def __new__(
        cls, kind: Tuple[Type[T], Type[T1], Type[T2]]
    ) -> Event[Union[T, T1, T2]]: ...
    @overload
    def __new__(cls, kind: Member[Any, T]) -> Event[T]: ...



# --------------------------------------------------------------------------------------
# Copyright (c) 2013-2023, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
from .catom import Validate
from .list import List


class ContainerList(List):
    """A List member which supports container notifications."""

    __slots__ = ()

    def __init__(self, item=None, default=None):
        """Initialize a ContainerList."""
        super(ContainerList, self).__init__(item, default)
        self.set_validate_mode(Validate.ContainerList, self.item)



# --------------------------------------------------------------------------------------
# Copyright (c) 2013-2023, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
from .catom import DefaultValue, Member, Validate
from .instance import Instance
from .typing_utils import extract_types, is_optional


class Tuple(Member):
    """A member which allows tuple values.

    If item validation is used, then assignment will create a copy of
    the original tuple before validating the items, since validation
    may change the item values.

    """

    __slots__ = ("item",)

    def __init__(self, item=None, default=()):
        """Initialize a Tuple.

        Parameters
        ----------
        item : Member, type, or tuple of types, optional
            A member to use for validating the types of items allowed in
            the tuple. This can also be a type object or a tuple of types,
            in which case it will be wrapped with an Instance member. If
            this is not given, no item validation is performed.

        default : tuple, optional
            The default tuple of values.

        """
        if item is not None and not isinstance(item, Member):
            opt, types = is_optional(extract_types(item))
            item = Instance(types, optional=opt)
        self.item = item
        self.set_default_value_mode(DefaultValue.Static, default)
        self.set_validate_mode(Validate.Tuple, item)

    def set_name(self, name):
        """Set the name of the member.

        This method ensures that the item member name is also updated.

        """
        super(Tuple, self).set_name(name)
        if self.item is not None:
            self.item.set_name(name + "|item")

    def set_index(self, index):
        """Assign the index to this member.

        This method ensures that the item member index is also updated.

        """
        super(Tuple, self).set_index(index)
        if self.item is not None:
            self.item.set_index(index)

    def clone(self):
        """Create a clone of the tuple.

        This will clone the internal tuple item if one is in use.

        """
        clone = super(Tuple, self).clone()
        item = self.item
        if item is not None:
            clone.item = item_clone = item.clone()
            mode, ctxt = self.validate_mode
            clone.set_validate_mode(mode, item_clone)
        else:
            clone.item = None
        return clone



# --------------------------------------------------------------------------------------
# Copyright (c) 2021-2023, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
import sys
from itertools import chain
from typing import (
    TYPE_CHECKING,
    Any,
    List,
    Literal,
    Sequence,
    Tuple,
    TypedDict,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

# In Python 3.9+, List is a _SpecialGenericAlias and does not inherit from
# _GenericAlias which is the type of List[int] for example
if sys.version_info >= (3, 9):
    from types import GenericAlias

    GENERICS = (type(List), type(List[int]), GenericAlias)
else:
    GENERICS = (type(List), type(List[int]))

if sys.version_info >= (3, 10):
    from types import UnionType

    UNION = (UnionType,)
else:
    UNION = ()

# Type checker consider that typing.List and typing.List[int] are types even though
# there are not at runtime.
if sys.version_info >= (3, 10):
    from types import GenericAlias, UnionType

    TypeLike = Union[type, TypeVar, UnionType, GenericAlias]
elif sys.version_info >= (3, 9):
    from types import GenericAlias

    TypeLike = Union[type, TypeVar, GenericAlias]
else:  # 3.8
    TypeLike = Union[type, TypeVar]

if TYPE_CHECKING:
    from .atom import Atom


class _ChangeDict(TypedDict):
    type: Union[
        Literal["create"],
        Literal["update"],
        Literal["delete"],
        Literal["event"],
        Literal["property"],
        Literal["container"],
    ]
    name: str
    object: "Atom"
    value: Any


class ChangeDict(_ChangeDict, total=False):
    oldvalue: Any

    # ContainerList specific entries, present only when type == "container"
    operation: Union[
        Literal["reverse"],
        Literal["__delitem__"],
        Literal["__iadd__"],
        Literal["__imul__"],
        Literal["__setitem__"],
        Literal["append"],
        Literal["extend"],
        Literal["insert"],
        Literal["pop"],
        Literal["remove"],
        Literal["sort"],
    ]
    # The following are present based on the operation value
    olditem: Any  # operation in ("__setitem__",)
    newitem: Any  # operation in ("__setitem__",)
    item: Any  # operation in ("append", "insert", "pop", "remove", "__delitem__")
    index: int  # operation in ("insert", "pop")
    items: Sequence[Any]  # operation in ("extend", "__iadd__")
    count: int  # operation in ("__imul__")
    key: Any  # operation in ("sort")
    reverse: bool  # operation in ("sort")


def _extract_types(kind: TypeLike) -> Tuple[type, ...]:
    """Extract a tuple of types from a type-like object"""
    if isinstance(kind, str):
        raise TypeError(
            f"Str-based annotations ({kind!r}) are not supported in atom Members."
        )

    ret: List[Any]
    if isinstance(kind, GENERICS):
        args = get_args(kind)
        origin = get_origin(kind)
        if origin is Union:
            ret = list(chain.from_iterable(extract_types(a) for a in args))
        else:
            ret = [origin]
    elif UNION and isinstance(kind, UNION):
        ret = list(chain.from_iterable(extract_types(a) for a in get_args(kind)))
    else:
        ret = [kind]

    extracted: List[type] = []
    for t in ret:
        if isinstance(t, TypeVar):
            b = t.__bound__
            if b is not None:
                if isinstance(b, str):
                    raise ValueError(
                        "Forward reference in type var bounds are not supported."
                    )
                assert isinstance(b, type)
                extracted.append(b)
            elif t.__constraints__:
                raise ValueError("Constraints in type var are not supported.")
            else:
                extracted.append(object)

            if t.__contravariant__:
                raise ValueError("TypeVar used in Atom object cannot be contravariant")

        elif t is Any:
            extracted.append(object)
        else:
            assert isinstance(t, type)
            extracted.append(t)

    return tuple(extracted)


def extract_types(kind: Union[TypeLike, Tuple[TypeLike, ...]]) -> Tuple[type, ...]:
    """Extract a tuple of types from a type-like object or tuple."""
    return tuple(
        chain.from_iterable(
            _extract_types(k) for k in (kind if isinstance(kind, tuple) else (kind,))
        )
    )


NONE_TYPE = type(None)


def is_optional(kinds: Tuple[type, ...]) -> Tuple[bool, Tuple[type, ...]]:
    """Determine if a tuple of types contains NoneType."""
    if NONE_TYPE in kinds:
        return True, tuple(k for k in kinds if k is not NONE_TYPE)
    else:
        return False, kinds



# --------------------------------------------------------------------------------------
# Copyright (c) 2021, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
from typing import Callable, NoReturn, Optional, TypeVar, overload

from .atom import Atom
from .catom import Member

A = TypeVar("A", bound=Atom)
T = TypeVar("T")
S = TypeVar("S")

class Property(Member[T, S]):
    @overload
    def __new__(
        cls,
        fget: None = None,
        fset: None = None,
        fdel: Optional[Callable[[A], None]] = None,
        cached: bool = False,
    ) -> Property[NoReturn, NoReturn]: ...
    @overload
    def __new__(
        cls,
        fget: None,
        fset: Callable[[A, S], None],
        fdel: Optional[Callable[[A], None]] = None,
        cached: bool = False,
    ) -> Property[NoReturn, S]: ...
    @overload
    def __new__(
        cls,
        fget: Callable[[A], T],
        fset: None = None,
        fdel: Optional[Callable[[A], None]] = None,
        cached: bool = False,
    ) -> Property[T, NoReturn]: ...
    @overload
    def __new__(
        cls,
        fget: Callable[[A], T],
        fset: Callable[[A, S], None],
        fdel: Optional[Callable[[A], None]] = None,
        cached: bool = False,
    ) -> Property[T, S]: ...
    @property
    def fget(self) -> Optional[Callable[[Atom], T]]: ...
    @property
    def fset(self) -> Optional[Callable[[Atom, S], None]]: ...
    @property
    def fdel(self) -> Optional[Callable[[Atom], None]]: ...
    @property
    def cached(self) -> bool: ...
    def getter(self, func: Callable[[A], T]) -> Callable[[A], T]: ...
    def setter(self, func: Callable[[A, S], None]) -> Callable[[A, S], None]: ...
    def deleter(self, func: Callable[[A], None]) -> Callable[[A], None]: ...
    def reset(self, owner: Atom) -> None: ...

def cached_property(fget: Callable[[A], T]) -> Property[T, NoReturn]: ...
